@@ -34,9 +34,40 @@
  * them. Pinned tabs always stay visible and aren't a part of the scrollable
  * area.
  *
+ * ## Drag-and-Drop
+ *
+ * `AdwTabBar` tabs can have an additional drop target for arbitrary content.
+ *
+ * Use [method@TabBar.setup_extra_drop_target] to set it up, specifying the
+ * supported content types and drag actions, then connect to
+ * [signal@TabBar::extra-drag-drop] to handle a drop.
+ *
+ * In some cases, it may be necessary to determine the used action based on the
+ * content. In that case, set [property@TabBar:extra-drag-preload] to `TRUE`
+ * and connect to [signal@TabBar::extra-drag-value] signal, then return the
+ * action from its handler. To access this action from the
+ * [signal@TabBar::extra-drag-drop] handler, use the
+ * [property@TabBar:extra-drag-preferred-action] property.
+ *
+ * [signal@TabBar::extra-drag-value] is also always emitted when starting to
+ * hover an item, with a `NULL` value. This happens even when
+ * [property@TabBar:extra-drag-preload] is `FALSE`.
+ *
  * ## CSS nodes
  *
  * `AdwTabBar` has a single CSS node with name `tabbar`.
+ *
+ * ## Style classes
+ *
+ * By default `AdwTabBar` look like a part of an `AdwHeaderBar` and is intended
+ * to be used directly attached to one or used as a [class@ToolbarView] toolbar.
+ * The [`.inline`](style-classes.html#inline) style class removes its background,
+ * so that it can be used in different contexts instead.
+ *
+ * <picture>
+ *   <source srcset="tab-bar-inline-dark.png" media="(prefers-color-scheme: dark)">
+ *   <img src="tab-bar-inline.png" alt="tab-bar-inline">
+ * </picture>
  */
 
 struct _AdwTabBar
@@ -291,6 +322,44 @@ view_destroy_cb (AdwTabBar *self)
   adw_tab_bar_set_view (self, NULL);
 }
 
+static gboolean
+handle_action_bin_focus (AdwTabBar        *self,
+                         AdwBin           *focused_bin,
+                         GtkDirectionType  dir,
+                         gboolean          dir_towards_tabs)
+{
+  if (gtk_widget_child_focus (GTK_WIDGET (focused_bin), dir))
+    return GDK_EVENT_STOP;
+
+  if (dir == GTK_DIR_UP || dir == GTK_DIR_DOWN)
+    return GDK_EVENT_PROPAGATE;
+
+  if (dir_towards_tabs) {
+    AdwTabPage *selected_page = adw_tab_view_get_selected_page (self->view);
+
+    if (!selected_page) {
+      GtkWidget *other_bin;
+
+      if (focused_bin == self->start_action_bin)
+        other_bin = GTK_WIDGET (self->end_action_bin);
+      else
+        other_bin = GTK_WIDGET (self->start_action_bin);
+
+      if (gtk_widget_child_focus (other_bin, dir))
+        return GDK_EVENT_STOP;
+
+      return gtk_widget_keynav_failed (GTK_WIDGET (self), dir);
+    }
+
+    if (adw_tab_page_get_pinned (selected_page))
+      return gtk_widget_child_focus (GTK_WIDGET (self->pinned_box), dir);
+
+    return gtk_widget_child_focus (GTK_WIDGET (self->box), dir);
+  }
+
+  return gtk_widget_keynav_failed (GTK_WIDGET (self), dir);
+}
+
 static gboolean
 adw_tab_bar_focus (GtkWidget        *widget,
                    GtkDirectionType  direction)
@@ -302,23 +371,65 @@ adw_tab_bar_focus (GtkWidget        *widget,
   if (!adw_tab_bar_get_tabs_revealed (self))
     return GDK_EVENT_PROPAGATE;
 
-  if (!gtk_widget_get_focus_child (widget))
-    return gtk_widget_child_focus (GTK_WIDGET (self->pinned_box), direction) ||
-           gtk_widget_child_focus (GTK_WIDGET (self->box), direction);
-
   is_rtl = gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL;
   start = is_rtl ? GTK_DIR_RIGHT : GTK_DIR_LEFT;
   end = is_rtl ? GTK_DIR_LEFT : GTK_DIR_RIGHT;
 
+  if (!gtk_widget_get_focus_child (widget)) {
+    if (direction == start || direction == GTK_DIR_TAB_BACKWARD || direction == GTK_DIR_UP) {
+      return gtk_widget_child_focus (GTK_WIDGET (self->end_action_bin), direction) ||
+             gtk_widget_child_focus (GTK_WIDGET (self->box), direction) ||
+             gtk_widget_child_focus (GTK_WIDGET (self->pinned_box), direction) ||
+             gtk_widget_child_focus (GTK_WIDGET (self->start_action_bin), direction);
+    } else {
+      return gtk_widget_child_focus (GTK_WIDGET (self->start_action_bin), direction) ||
+             gtk_widget_child_focus (GTK_WIDGET (self->pinned_box), direction) ||
+             gtk_widget_child_focus (GTK_WIDGET (self->box), direction) ||
+             gtk_widget_child_focus (GTK_WIDGET (self->end_action_bin), direction);
+    }
+  }
+
+  if (gtk_widget_get_focus_child (GTK_WIDGET (self->start_action_bin))) {
+    return handle_action_bin_focus (self,
+                                    self->start_action_bin,
+                                    direction,
+                                    direction == end || direction == GTK_DIR_TAB_FORWARD);
+  }
+
+  if (gtk_widget_get_focus_child (GTK_WIDGET (self->end_action_bin))) {
+    return handle_action_bin_focus (self,
+                                    self->end_action_bin,
+                                    direction,
+                                    direction == start || direction == GTK_DIR_TAB_BACKWARD);
+  }
+
+  /* If the focus is not in either action_bin, then it must be in the tabs */
+
   if (direction == start) {
-    if (adw_tab_view_select_previous_page (self->view))
+    if (adw_tab_view_select_previous_page (self->view) ||
+        gtk_widget_child_focus (GTK_WIDGET (self->start_action_bin), direction))
       return GDK_EVENT_STOP;
 
     return gtk_widget_keynav_failed (widget, direction);
   }
 
   if (direction == end) {
-    if (adw_tab_view_select_next_page (self->view))
+    if (adw_tab_view_select_next_page (self->view) ||
+        gtk_widget_child_focus (GTK_WIDGET (self->end_action_bin), direction))
+      return GDK_EVENT_STOP;
+
+    return gtk_widget_keynav_failed (widget, direction);
+  }
+
+  if (direction == GTK_DIR_TAB_BACKWARD) {
+    if (gtk_widget_child_focus (GTK_WIDGET (self->start_action_bin), direction))
+      return GDK_EVENT_STOP;
+
+    return gtk_widget_keynav_failed (widget, direction);
+  }
+
+  if (direction == GTK_DIR_TAB_FORWARD) {
+    if (gtk_widget_child_focus (GTK_WIDGET (self->end_action_bin), direction))
       return GDK_EVENT_STOP;
 
     return gtk_widget_keynav_failed (widget, direction);
@@ -546,11 +657,12 @@ adw_tab_bar_class_init (AdwTabBarClass *klass)
   /**
    * AdwTabBar:extra-drag-preferred-action:
    *
-   * The unique action on the `current-drop` of the
-   * [signal@TabBar::extra-drag-drop].
+   * The current drag action during a drop.
    *
-   * This property should only be used during a [signal@TabBar::extra-drag-drop]
-   * and is always a subset of what was originally passed to
+   * This property should only be used from inside a
+   * [signal@TabBar::extra-drag-drop] handler.
+   *
+   * The action will be a subset of what was originally passed to
    * [method@TabBar.setup_extra_drop_target].
    *
    * Since: 1.4
@@ -582,14 +694,14 @@ adw_tab_bar_class_init (AdwTabBarClass *klass)
    * @page: the page matching the tab the content was dropped onto
    * @value: the `GValue` being dropped
    *
-   * This signal is emitted when content is dropped onto a tab.
+   * Emitted when content is dropped onto a tab.
    *
    * The content must be of one of the types set up via
    * [method@TabBar.setup_extra_drop_target].
    *
    * See [signal@Gtk.DropTarget::drop].
    *
-   * Returns: whether the drop was accepted for @page
+   * Returns: whether the drop was accepted
    */
   signals[SIGNAL_EXTRA_DRAG_DROP] =
     g_signal_new ("extra-drag-drop",
@@ -607,9 +719,9 @@ adw_tab_bar_class_init (AdwTabBarClass *klass)
    * AdwTabBar::extra-drag-value:
    * @self: a tab bar
    * @page: the page matching the tab the content was dropped onto
-   * @value: the `GValue` being dropped
+   * @value: (nullable): the `GValue` being dropped
    *
-   * This signal is emitted when the dropped content is preloaded.
+   * Emitted when the dropped content is preloaded.
    *
    * In order for data to be preloaded, [property@TabBar:extra-drag-preload]
    * must be set to `TRUE`.
@@ -619,7 +731,7 @@ adw_tab_bar_class_init (AdwTabBarClass *klass)
    *
    * See [property@Gtk.DropTarget:value].
    *
-   * Returns: the preferred action for the drop on @page
+   * Returns: the preferred action for the drop
    *
    * Since: 1.3
    */
@@ -1079,8 +1191,6 @@ adw_tab_bar_set_inverted (AdwTabBar *self,
  *   all supported `GType`s that can be dropped
  * @n_types: number of @types
  *
- * Sets the supported types for this drop target.
- *
  * Sets up an extra drop target on tabs.
  *
  * This allows to drag arbitrary content onto tabs, for example URLs in a web
@@ -1108,9 +1218,15 @@ adw_tab_bar_setup_extra_drop_target (AdwTabBar     *self,
  * adw_tab_bar_get_extra_drag_preferred_action:
  * @self: a tab bar
  *
- * Gets the current action during a drop on the extra_drop_target.
+ * Gets the current drag action during a drop.
  *
- * Returns: the drag action of the current drop.
+ * This method should only be used from inside a
+ * [signal@TabBar::extra-drag-drop] handler.
+ *
+ * The action will be a subset of what was originally passed to
+ * [method@TabBar.setup_extra_drop_target].
+ *
+ * Returns: the drag action of the current drop
  *
  * Since: 1.4
  */
